@@ -32,80 +32,99 @@
 
 #include "log.h"
 #include "gw-obex.h"
+#include "utils.h"
+#include "obex-xfer.h"
 #include "obex-priv.h"
 
-#define CHECK_DISCONNECT if (ctx->conn_fd < 0) { \
-                            if (error) \
-                                *error = GW_OBEX_ERROR_DISCONNECT; \
-                            GW_OBEX_UNLOCK(ctx); \
-                            return FALSE; \
-                         }
 
-gboolean gw_obex_get_file(GwObex *ctx, const gchar *local, const gchar *remote, gint *error) {
+gboolean gw_obex_get_file(GwObex *ctx,
+                          const gchar *local,
+                          const gchar *remote,
+                          const gchar *type,
+                          gint *error) {
     gboolean ret;
     GW_OBEX_LOCK(ctx);
-    CHECK_DISCONNECT
-    ret = gw_obex_get(ctx, local, remote, NULL, NULL, NULL);
-    if (ret == FALSE && error)
+    CHECK_DISCONNECT(FALSE, error, ctx);
+    ret = gw_obex_get(ctx, local, remote, type, NULL, NULL, -1, FALSE);
+    if (ret == FALSE)
         gw_obex_get_error(ctx, error);
     GW_OBEX_UNLOCK(ctx);
     return ret;
 }
 
-gboolean gw_obex_get_fd(GwObex *ctx, gint fd, const gchar *remote, gint *error) {
+gboolean gw_obex_get_fd(GwObex      *ctx,
+                        gint         fd,
+                        const gchar *remote,
+                        const gchar *type,
+                        gint        *error) {
     gboolean ret;
     GW_OBEX_LOCK(ctx);
-    CHECK_DISCONNECT
-    ctx->stream_fd = fd;
-    ret = gw_obex_get(ctx, NULL, remote, NULL, NULL, NULL);
-    if (ret == FALSE && error)
+    CHECK_DISCONNECT(FALSE, error, ctx);
+    ret = gw_obex_get(ctx, NULL, remote, type, NULL, NULL, fd, FALSE);
+    if (ret == FALSE)
         gw_obex_get_error(ctx, error);
     GW_OBEX_UNLOCK(ctx);
     return ret;
 }
 
-gboolean gw_obex_put_fd(GwObex *ctx, gint fd, const gchar *remote, gint *error) {
+gboolean gw_obex_put_fd(GwObex      *ctx,
+                        gint         fd,
+                        const gchar *remote,
+                        const gchar *type,
+                        gint        *error) {
     gboolean ret;
     GW_OBEX_LOCK(ctx);
-    CHECK_DISCONNECT
-    ctx->stream_fd = fd;
-    ret = gw_obex_put(ctx, NULL, remote, NULL, NULL, 0);
-    if (ret == FALSE && error)
+    CHECK_DISCONNECT(FALSE, error, ctx);
+    ret = gw_obex_put(ctx, NULL, remote, type, NULL, 0, -1, fd, FALSE);
+    if (ret == FALSE)
         gw_obex_get_error(ctx, error);
     GW_OBEX_UNLOCK(ctx);
     return ret;
 }
 
-gboolean gw_obex_put_file(GwObex *ctx, const gchar *local, const gchar *remote, gint *error) {
+gboolean gw_obex_put_file(GwObex      *ctx,
+                          const gchar *local,
+                          const gchar *remote,
+                          const gchar *type,
+                          gint        *error) {
     gboolean ret;
     GW_OBEX_LOCK(ctx);
-    CHECK_DISCONNECT
-    ret = gw_obex_put(ctx, local, remote, NULL, NULL, 0);
-    if (ret == FALSE && error)
+    CHECK_DISCONNECT(FALSE, error, ctx);
+    ret = gw_obex_put(ctx, local, remote, type, NULL, 0, -1, -1, FALSE);
+    if (ret == FALSE)
         gw_obex_get_error(ctx, error);
     GW_OBEX_UNLOCK(ctx);
     return ret;
 }
 
-gboolean gw_obex_get_buf(GwObex *ctx, const gchar *remote, const gchar *type,
-                         gchar **buf, gint *buf_size, gint *error) {
+gboolean gw_obex_get_buf(GwObex       *ctx,
+                         const gchar  *remote,
+                         const gchar  *type,
+                         gchar       **buf,
+                         gint         *buf_size,
+                         gint         *error) {
     gboolean ret;
     GW_OBEX_LOCK(ctx);
-    CHECK_DISCONNECT
-    ret = gw_obex_get(ctx, NULL, remote, type, buf, buf_size);
-    if (ret == FALSE && error)
+    CHECK_DISCONNECT(FALSE, error, ctx);
+    ret = gw_obex_get(ctx, NULL, remote, type, buf, buf_size, -1, FALSE);
+    if (ret == FALSE)
         gw_obex_get_error(ctx, error);
     GW_OBEX_UNLOCK(ctx);
     return ret;
 }
 
-gboolean gw_obex_put_buf(GwObex *ctx, const gchar *remote, const gchar *type,
-                         const gchar *buf, gint buf_size, gint *error) {
+gboolean gw_obex_put_buf(GwObex      *ctx,
+                         const gchar *remote,
+                         const gchar *type,
+                         const gchar *buf,
+                         gint         buf_size,
+                         gint         time,       
+                         gint        *error) {
     gboolean ret;
     GW_OBEX_LOCK(ctx);
-    CHECK_DISCONNECT
-    ret = gw_obex_put(ctx, NULL, remote, type, buf, buf_size);
-    if (ret == FALSE && error)
+    CHECK_DISCONNECT(FALSE, error, ctx);
+    ret = gw_obex_put(ctx, NULL, remote, type, buf, buf_size, time, -1, FALSE);
+    if (ret == FALSE)
         gw_obex_get_error(ctx, error);
     GW_OBEX_UNLOCK(ctx);
     return ret;
@@ -114,9 +133,9 @@ gboolean gw_obex_put_buf(GwObex *ctx, const gchar *remote, const gchar *type,
 gboolean gw_obex_chdir(GwObex *ctx, const gchar *dir, gint *error) {
     gboolean ret;
     GW_OBEX_LOCK(ctx);
-    CHECK_DISCONNECT
-    ret = gw_obex_setpath(ctx, dir, 0);
-    if (ret == FALSE && error)
+    CHECK_DISCONNECT(FALSE, error, ctx);
+    ret = gw_obex_setpath(ctx, dir ? dir : "", 0);
+    if (ret == FALSE)
         gw_obex_get_error(ctx, error);
     GW_OBEX_UNLOCK(ctx);
     return ret;
@@ -124,10 +143,9 @@ gboolean gw_obex_chdir(GwObex *ctx, const gchar *dir, gint *error) {
 
 gboolean gw_obex_mkdir(GwObex *ctx, const gchar *dir, gint *error) {
     GW_OBEX_LOCK(ctx);
-    CHECK_DISCONNECT
-    if (!gw_obex_setpath(ctx, dir, SETPATH_CREATE)) {
-        if (error)
-            gw_obex_get_error(ctx, error);
+    CHECK_DISCONNECT(FALSE, error, ctx);
+    if (!gw_obex_setpath(ctx, dir ? dir : "", SETPATH_CREATE)) {
+        gw_obex_get_error(ctx, error);
         GW_OBEX_UNLOCK(ctx);
         return FALSE;
     }
@@ -140,12 +158,10 @@ gboolean gw_obex_read_dir(GwObex *ctx, const gchar *dir,
                           gchar **buf, gint *buf_size, gint *error) {
     gboolean ret;
     GW_OBEX_LOCK(ctx);
-    CHECK_DISCONNECT
-    ret = gw_obex_get(ctx, NULL, dir ? dir : "", LST_TYPE, buf, buf_size);
-    if (ret == FALSE) {
-        if (error)
-            gw_obex_get_error(ctx, error);
-    }
+    CHECK_DISCONNECT(FALSE, error, ctx);
+    ret = gw_obex_get(ctx, NULL, dir ? dir : "", LST_TYPE, buf, buf_size, -1, FALSE);
+    if (ret == FALSE)
+        gw_obex_get_error(ctx, error);
     else if (*buf_size > 0) {
         /* Hack for some OBEX implementations which send nul's
          * at the end of the listing */
@@ -165,9 +181,9 @@ gboolean gw_obex_read_dir(GwObex *ctx, const gchar *dir,
 gboolean gw_obex_delete(GwObex *ctx, const gchar *name, gint *error) {
     gboolean ret;
     GW_OBEX_LOCK(ctx);
-    CHECK_DISCONNECT
-    ret = gw_obex_put(ctx, NULL, name, NULL, NULL, 0);
-    if (ret == FALSE && error)
+    CHECK_DISCONNECT(FALSE, error, ctx);
+    ret = gw_obex_put(ctx, NULL, name, NULL, NULL, 0, -1, -1, FALSE);
+    if (ret == FALSE)
         gw_obex_get_error(ctx, error);
     GW_OBEX_UNLOCK(ctx);
     return ret;
@@ -176,9 +192,9 @@ gboolean gw_obex_delete(GwObex *ctx, const gchar *name, gint *error) {
 gboolean gw_obex_move(GwObex *ctx, const gchar *src, const gchar *dst, gint *error) {
     gboolean ret;
     GW_OBEX_LOCK(ctx);
-    CHECK_DISCONNECT
+    CHECK_DISCONNECT(FALSE, error, ctx);
     ret = gw_obex_action_op(ctx, src, dst, OBEX_ACTION_MOVE);
-    if (ret == FALSE && error)
+    if (ret == FALSE)
         gw_obex_get_error(ctx, error);
     GW_OBEX_UNLOCK(ctx);
     return ret;
@@ -187,9 +203,9 @@ gboolean gw_obex_move(GwObex *ctx, const gchar *src, const gchar *dst, gint *err
 gboolean gw_obex_copy(GwObex *ctx, const gchar *src, const gchar *dst, gint *error) {
     gboolean ret;
     GW_OBEX_LOCK(ctx);
-    CHECK_DISCONNECT
+    CHECK_DISCONNECT(FALSE, error, ctx);
     ret = gw_obex_action_op(ctx, src, dst, OBEX_ACTION_COPY);
-    if (ret == FALSE && error)
+    if (ret == FALSE)
         gw_obex_get_error(ctx, error);
     GW_OBEX_UNLOCK(ctx);
     return ret;
@@ -198,29 +214,12 @@ gboolean gw_obex_copy(GwObex *ctx, const gchar *src, const gchar *dst, gint *err
 gboolean gw_obex_get_capability(GwObex *ctx, gchar **cap, gint *cap_len, gint *error) {
     gboolean ret;
     GW_OBEX_LOCK(ctx);
-    CHECK_DISCONNECT
-    ret = gw_obex_get(ctx, NULL, NULL, CAP_TYPE, cap, cap_len);
+    CHECK_DISCONNECT(FALSE, error, ctx);
+    ret = gw_obex_get(ctx, NULL, NULL, CAP_TYPE, cap, cap_len, -1, FALSE);
     if (ret == FALSE) {
         *cap = NULL;
         *cap_len = 0;
-        if (error)
-            gw_obex_get_error(ctx, error);
-    }
-    GW_OBEX_UNLOCK(ctx);
-    return ret;
-}
-
-gboolean gw_obex_get_type(GwObex *ctx, const gchar *type, gchar **buf, gint *buf_len, gint *error) {
-    gboolean ret;
-    g_return_val_if_fail(type != NULL, FALSE);
-    GW_OBEX_LOCK(ctx);
-    CHECK_DISCONNECT
-    ret = gw_obex_get(ctx, NULL, NULL, type, buf, buf_len);
-    if (ret == FALSE) {
-        *buf = NULL;
-        *buf_len = 0;
-        if (error)
-            gw_obex_get_error(ctx, error);
+        gw_obex_get_error(ctx, error);
     }
     GW_OBEX_UNLOCK(ctx);
     return ret;
@@ -249,6 +248,16 @@ void gw_obex_set_cancel_callback(GwObex *ctx, gw_obex_cancel_cb_t callback, gpoi
 
 void gw_obex_close(GwObex *ctx) {
     GW_OBEX_LOCK(ctx);
+    if (ctx->xfer) {
+        GwObexXfer *xfer = ctx->xfer;
+        GW_OBEX_UNLOCK(ctx);
+        gw_obex_xfer_close(ctx->xfer, NULL);
+        GW_OBEX_LOCK(ctx);
+        /* In the async case the caller of put/get_async owns the xfer object */
+        if (!xfer->async)
+            _gw_obex_xfer_free(xfer);
+        ctx->xfer = NULL;
+    }
     if (ctx->conn_fd >= 0) {
         if (!gw_obex_disconnect(ctx))
             debug("OBEX Disconnect command failed\n");
@@ -268,11 +277,6 @@ void gw_obex_close(GwObex *ctx) {
         g_source_destroy(ctx->gio_source);
         ctx->gio_source = NULL;
     }
-    if (ctx->buf) {
-        g_free(ctx->buf);
-        ctx->buf = NULL;
-        ctx->buf_size = 0;
-    }
     GW_OBEX_UNLOCK(ctx);
 #ifdef G_THREADS_ENABLED
     g_mutex_free(ctx->mutex);
@@ -319,10 +323,12 @@ GwObex *gw_obex_setup_fd(int fd, const gchar *uuid, gint uuid_len,
 
     ctx->gio = g_io_channel_unix_new(ctx->conn_fd);
     ctx->gio_source = g_io_create_watch (ctx->gio,
-                                         G_IO_HUP | G_IO_ERR | G_IO_NVAL);
+                                         G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL);
     g_source_set_callback(ctx->gio_source, (GSourceFunc)gw_obex_cb, ctx, NULL);
     (void) g_source_attach(ctx->gio_source, context);
 
+    ctx->main_ctx = context;
+
     return ctx;
 }
 
@@ -340,7 +346,7 @@ GwObex *gw_obex_setup_dev(const char *dev, const gchar *uuid, gint uuid_len,
     }
 
     if (!fd_raw_mode(fd)) {
-        debug("setting raw mode failed");
+        debug("setting raw mode failed\n");
         close(fd);
         if (error)
             *error = GW_OBEX_ERROR_CONNECT_FAILED;
