@@ -64,14 +64,18 @@
 /** Local access error (e.g. read/write/open failed for local file) */
 #define GW_OBEX_ERROR_LOCAL_ACCESS      264
 
+/** Another operation is in progress */
+#define GW_OBEX_ERROR_BUSY              265
+
+/** No data currently available */
+#define GW_OBEX_ERROR_NO_DATA           266
+
+
 /** @} */
 
 /** Value used if target length for put or get is not known */
 #define GW_OBEX_UNKNOWN_LENGTH -1
 
-/** Struct containing the context of a gwobex connection */
-typedef struct gw_obex GwObex;
-
 /** Standard folder browsing service UUID (give this as a parameter to
  *  gw_obex_setup_* to connect to folder browsing service */
 #define OBEX_FTP_UUID \
@@ -79,6 +83,25 @@ typedef struct gw_obex GwObex;
 /** Length of OBEX_FTP_UUID */
 #define OBEX_FTP_UUID_LEN 16
 
+/** Phone Book Access Profile UUID */
+#define OBEX_PBAP_UUID \
+    "\x79\x61\x35\xF0\xF0\xC5\x11\xD8\x09\x66\x08\x00\x20\x0C\x9A\x66"
+/** Length of OBEX_PBAP_UUID */
+#define OBEX_PBAP_UUID_LEN 16
+
+/** Struct containing the context of a gwobex connection */
+typedef struct gw_obex GwObex;
+
+/** Objecct transfer handle */
+typedef struct gw_obex_xfer GwObexXfer;
+
+/** Callback type for ongoing transfers
+ * @param ctx  GwObexXfer pointer for the transfer
+ * @param data Optional pointer to user data
+ */
+typedef void (*gw_obex_xfer_cb_t) (GwObexXfer *xfer,
+                                   gpointer data);
+
 /** Callback type for transport connection loss
  * @param ctx  GwObex pointer for the connection
  * @param data Optional pointer to user data
@@ -106,6 +129,14 @@ typedef void (*gw_obex_progress_cb_t) (GwObex *ctx, gint obex_cmd,
 typedef gboolean (*gw_obex_cancel_cb_t) (gpointer data);
 
 
+/**
+ * @name Functions for connecting and disconnecting
+ * With these functions you can create and and disconnect connections. You can
+ * either connect using a filename (e.g. "/dev/rfcomm0") or using a file
+ * descriptor (e.g. a RFCOMM socket).
+ * @{
+ */
+
 /** Open connection using a local device node and setup parameters.
  * This function should be called before calling any other functions. The
  * pointer returned by this function should be passed to the other functions.
@@ -127,11 +158,11 @@ GwObex *gw_obex_setup_dev(const gchar *device,
                           gint *error);
 
 
-/** Setup OBEX connection using an opened filedescriptor
+/** Setup OBEX connection using an opened file descriptor
  * This function should be called before calling any other functions. The
  * pointer returned by this function should be passed to the other functions.
  *
- * @param fd       Opened filedescriptor to use for the connection
+ * @param fd       Opened file descriptor to use for the connection
  * @param uuid     UUID of service to connect to. NULL for the default service
  *                 (INBOX).
  * @param uuid_len Length (in bytes) of UUID
@@ -155,42 +186,14 @@ GwObex *gw_obex_setup_fd(int fd,
  */
 void gw_obex_close(GwObex *ctx);
 
+/** @} */
 
-/** Get the capability object from the connected remote device.
- *
- * @param ctx     Pointer returned by gw_obex_setup()
- * @param cap     Place to store the fetched object.
- *                 g_free() when not needed anymore.
- * @param cap_len Place to store the size of the fetched object
- * @param error   Place to store a possible error code
- *   (NULL if not interested)
- *
- * @returns TRUE on success, FALSE on failure
+/**
+ * @name Registering callback functions
+ * With these functions you can register your own callback functions
+ * to gwobex to receive indications about special events.
+ * @{
  */
-gboolean gw_obex_get_capability(GwObex *ctx,
-                                gchar **cap,
-                                gint *cap_len,
-                                gint *error);
-
-
-/** Get an object identified by its type
- * 
- * @param ctx     Pointer returned by gw_obex_setup()
- * @param type    Type of the object to fetch
- * @param buf     Place to store the fetched object.
- *                 g_free() when not needed anymore.
- * @param buf_len Place to store the size of the fetched object
- * @param error   Place to store a possible error code
- *   (NULL if not interested)
- *
- * @returns TRUE on success, FALSE on failure
- */
-gboolean gw_obex_get_type(GwObex *ctx,
-                          const gchar *type,
-                          gchar **buf,
-                          gint *buf_len,
-                          gint *error);
-
 
 /** Set function to be called when a disconnection happens.
  *  You may (and probably should) call gw_obex_close() if this function is
@@ -228,12 +231,38 @@ void gw_obex_set_cancel_callback(GwObex *ctx,
                                  gw_obex_cancel_cb_t callback,
                                  gpointer data);
 
+/** @} */
+
+/**
+ * @name Functions for performing synchronous remote operations
+ * Once you have setup a connection using one of the gw_obex_setup_* functions,
+ * you can perform different remote transactions using these functions.
+ * @{
+ */
+
+/** Get the capability object from the connected remote device.
+ *
+ * @param ctx     Pointer returned by gw_obex_setup()
+ * @param cap     Place to store the fetched object.
+ *                 g_free() when not needed anymore.
+ * @param cap_len Place to store the size of the fetched object
+ * @param error   Place to store a possible error code
+ *   (NULL if not interested)
+ *
+ * @returns TRUE on success, FALSE on failure
+ */
+gboolean gw_obex_get_capability(GwObex *ctx,
+                                gchar **cap,
+                                gint *cap_len,
+                                gint *error);
+
 
 /** Get a file from the remote device.
  *
  * @param ctx    Pointer returned by gw_obex_setup()
  * @param local  Local filename (null terminated UTF-8)
  * @param remote Remote filename (null terminated UTF-8)
+ * @param type   MIME-type of the object
  * @param error  Place to store error code on failure
  *               (NULL if not interested)
  *
@@ -242,6 +271,7 @@ void gw_obex_set_cancel_callback(GwObex *ctx,
 gboolean gw_obex_get_file(GwObex *ctx,
                           const gchar *local,
                           const gchar *remote,
+                          const gchar *type,
                           gint *error);
 
 
@@ -250,6 +280,7 @@ gboolean gw_obex_get_file(GwObex *ctx,
  * @param ctx    Pointer returned by gw_obex_setup()
  * @param local  Local filename (null terminated UTF-8)
  * @param remote Remote filename (null terminated UTF-8)
+ * @param type   MIME-type of the object
  * @param error  Place to store error code on failure
  *               (NULL if not interested)
  *
@@ -258,6 +289,7 @@ gboolean gw_obex_get_file(GwObex *ctx,
 gboolean gw_obex_put_file(GwObex *ctx,
                           const gchar *local,
                           const gchar *remote,
+                          const gchar *type,
                           gint *error);
 
 
@@ -266,24 +298,32 @@ gboolean gw_obex_put_file(GwObex *ctx,
  * @param ctx    Pointer returned by gw_obex_setup()
  * @param fd     File descriptor to write the file into
  * @param remote Remote filename (null terminated UTF-8)
+ * @param type   MIME-type of the object
  * @param error  Place to store error code on failure
  *               (NULL if not interested)
  *
  * @returns TRUE on success, FALSE on failure
  */
-gboolean gw_obex_get_fd(GwObex *ctx, gint fd, const gchar *remote, gint *error);
+gboolean gw_obex_get_fd(GwObex *ctx, gint fd,
+                        const gchar *remote,
+                        const gchar *type,
+                        gint *error);
 
 /** Read data from a file descriptor and send it to the remote device
  *
  * @param ctx    Pointer returned by gw_obex_setup()
  * @param fd     File descriptor to read the data from 
  * @param remote Remote filename (null terminated UTF-8)
+ * @param type   MIME-type of the object
  * @param error  Place to store error code on failure
  *               (NULL if not interested)
  *
  * @returns TRUE on success, FALSE on failure
  */
-gboolean gw_obex_put_fd(GwObex *ctx, gint fd, const gchar *remote, gint *error);
+gboolean gw_obex_put_fd(GwObex *ctx, gint fd,
+                        const gchar *remote,
+                        const gchar *type,
+                        gint *error);
 
 /** Get an object from the remote device and store it in a memory buffer.
  * Either remote filename or type must be supplied (or both).
@@ -311,13 +351,14 @@ gboolean gw_obex_get_buf(GwObex *ctx, const gchar *remote, const gchar *type,
  * @param type     MIME-type of the object
  * @param buf      Buffer containing the object
  * @param buf_size Buffer (object) size
+ * @param time     Last modification time of object (or -1 if not known)
  * @param error    Place to store error code on failure
  *                 (NULL if not interested)
  *
  * @returns TRUE on success, FALSE on failure
  */
 gboolean gw_obex_put_buf(GwObex *ctx, const gchar *remote, const gchar *type,
-                         const gchar *buf, gint buf_size, gint *error);
+                         const gchar *buf, gint buf_size, gint time, gint *error);
 
 
 /** Change directory (relative to the current one).
@@ -394,6 +435,163 @@ gboolean gw_obex_move(GwObex *ctx, const gchar *src, const gchar *dst,
 gboolean gw_obex_copy(GwObex *ctx, const gchar *src, const gchar *dst,
                       gint *error);
 
+/** @} */
+
+/**
+ * @name Functions for performing transfers in an asynchronous manner
+ * With these functions you can do transfers in smaller steps. The steps
+ * are split up in a open, read/write, close manner.
+ * @{
+ */
+
+/** Start a PUT operation asynchronously
+ *
+ * @param ctx   Pointer returned by gw_obex_setup()
+ * @param name  Name of the object (null terminated UTF-8)
+ * @param type  Type of the object (null terminated UTF-8), or NULL
+ * @param size  Size of the object (GW_OBEX_UNKNOWN_LENGTH if not known)
+ * @param time  Last modification time of the object (-1 if not known)
+ * @param error Place to store error code on failure (NULL if not interested)
+ *
+ * @returns a new GwObexXfer object on success, NULL on failure
+ */
+GwObexXfer *gw_obex_put_async(GwObex *ctx, const char *name, const char *type,
+                              gint size, time_t time, gint *error);
+
+
+/** Start a GET operation asynchronously
+ *
+ * @param ctx   Pointer returned by gw_obex_setup()
+ * @param name  Name of the object (null terminated UTF-8)
+ * @param type  Type of the object (null terminated UTF-8), or NULL
+ * @param error Place to store error code on failure (NULL if not interested)
+ *
+ * @returns a new GwObexXfer object on success, NULL on failure
+ */
+GwObexXfer *gw_obex_get_async(GwObex *ctx, const char *name, const char *type, gint *error);
+
+
+/** Set a callback function for a GwObexXfer object
+ * The callback function will be called in the following situations:
+ * <ul>
+ *  <li>Data can be written (i.e. xfer_write will succeed)</li>
+ *  <li>Data can be read (i.e. xfer_read will succees)</li>
+ *  <li>An error ocured</li>
+ *  <li>The transfer is finished</li>
+ * </ul>
+ *
+ * @param xfer      Pointer returned by gw_obex_put_async or gw_obex_get_async
+ * @param cb        Pointer to the callback function
+ * @param user_data Optional user data which will be passed to the callback function
+ *
+ * @returns a new GwObexXfer object on success, NULL on failure
+ */
+void gw_obex_xfer_set_callback(GwObexXfer *xfer, gw_obex_xfer_cb_t cb, gpointer user_data);
+
+
+/** Get the last modification time of the object being transfered
+ *
+ * @param xfer Pointer returned by gw_obex_put_async or gw_obex_get_async
+ *
+ * @returns The modification time or -1 if it is not known.
+ */
+time_t gw_obex_xfer_object_time(GwObexXfer *xfer);
+
+
+/** Get the size of the object being transfered
+ *
+ * @param xfer Pointer returned by gw_obex_put_async or gw_obex_get_async
+ *
+ * @returns The size or GW_OBEX_UNKNOWN_LENGTH if it is not known.
+ */
+gint gw_obex_xfer_object_size(GwObexXfer *xfer);
+
+
+/** Supply more data to a transfer
+ *
+ * @param xfer          Pointer returned by gw_obex_put_async or gw_obex_get_async
+ * @param buf           Buffer containing the data
+ * @param buf_size      Size of the buffer
+ * @param bytes_written Return value for the number of bytes that were written
+ * @param error Place to store error code on failure (NULL if not interested)
+ *
+ * @returns TRUE on success, FALSE on failure
+ */
+gboolean gw_obex_xfer_write(GwObexXfer *xfer, const char *buf, gint buf_size,
+                            gint *bytes_written, gint *error);
+
+/** Read data from a transfer
+ *
+ * The function will report EOF by returning success with zero bytes read.
+ *
+ * @param xfer          Pointer returned by gw_obex_put_async or gw_obex_get_async
+ * @param buf           Buffer where the data should be stored
+ * @param buf_size      Size of the buffer
+ * @param bytes_read    Return value for the number of bytes that were read
+ * @param error Place to store error code on failure (NULL if not interested)
+ *
+ * @returns TRUE on success, FALSE on failure
+ */
+gboolean gw_obex_xfer_read(GwObexXfer *xfer, char *buf, gint buf_size,
+                           gint *bytes_read, gint *error);
+
+
+/** Force all data remaining in buffers to be sent
+ *
+ * @param xfer  Pointer returned by gw_obex_put_async or gw_obex_get_async
+ * @param error Place to store error code on failure (NULL if not interested)
+ *
+ * @returns TRUE on success, FALSE on failure
+ */
+gboolean gw_obex_xfer_flush(GwObexXfer *xfer, gint *error);
+
+
+/** Close an ongoing transfer
+ *
+ * You still need to call gw_obex_xfer_free after this to free the actual
+ * memory allocated for the GwObexXfer object.
+ *
+ * @param xfer  Pointer returned by gw_obex_put_async or gw_obex_get_async
+ * @param error Place to store error code on failure (NULL if not interested)
+ *
+ * @returns TRUE on success, FALSE on failure
+ */
+gboolean gw_obex_xfer_close(GwObexXfer *xfer, gint *error);
+
+
+/** Abort an ongoing transfer
+ *
+ * You still need to call gw_obex_xfer_free after this to free the actual
+ * memory allocated for the GwObexXfer object. xfer_close and xfer_abort are
+ * mutually exclusive (only call one of them for a transfer).
+ *
+ * @param xfer  Pointer returned by gw_obex_put_async or gw_obex_get_async
+ * @param error Place to store error code on failure (NULL if not interested)
+ *
+ * @returns TRUE on success, FALSE on failure
+ */
+gboolean gw_obex_xfer_abort(GwObexXfer *xfer, gint *error);
+
+
+/** Free the data allocated for a GwObexXfer object
+ *
+ * @param xfer  Pointer returned by gw_obex_put_async or gw_obex_get_async
+ */
+void gw_obex_xfer_free(struct gw_obex_xfer *xfer);
+
+
+/** Set blocking behaviour for a GwObexXfer object when calling xfer_read and xfer_write
+ *
+ * When blocking is enabled xfer_read will return only after it has been able
+ * to read some data (i.e. GW_OBEX_ERROR_NO_DATA will not be returned). For xfer_write
+ * blocking guarantees that *some* data will be written.
+ *
+ * @param xfer  Pointer returned by gw_obex_put_async or gw_obex_get_async
+ * @param block TRUE to enable blocking behaviour
+ */
+void gw_obex_xfer_set_blocking(GwObexXfer *xfer, gboolean block);
+
+/** @} */
 
 #endif /* _GW_OBEX_H_ */
 
