@@ -34,11 +34,7 @@
 #include <utime.h>
 #include <fcntl.h>
 #include <strings.h>
-#include <termios.h>
 #include <glib.h>
-#ifdef DEBUG
-#include <time.h>
-#endif
 
 #include <sys/types.h>
 #include <sys/socket.h>
@@ -51,8 +47,9 @@
 
 #include "log.h"
 #include "gw-obex.h"
-#include "obex-priv.h"
 #include "utils.h"
+#include "obex-xfer.h"
+#include "obex-priv.h"
 
 static gboolean file_is_dir(const char *filename) {
     struct stat st;
@@ -66,79 +63,34 @@ static gboolean file_is_dir(const char *filename) {
     return FALSE;
 }
 
-static void gw_obex_set_error(GwObex *ctx) {
-    if (ctx->abort)
-        ctx->error = GW_OBEX_ERROR_ABORT;
-    else if (ctx->conn_fd < 0 || ctx->link_err)
-        ctx->error = GW_OBEX_ERROR_DISCONNECT;
-    else
-        ctx->error = (gint)ctx->obex_rsp;
-}
-
 static void idle_callback(obex_t *handle, obex_object_t *object, int mode,
                     int event, int obex_cmd, int obex_rsp) {
     debug("idle_callback() called\n");
     sleep(1);
 }
 
-static void obex_link_error(GwObex *ctx) {
-    if (ctx->link_err)
-        return;
-    ctx->link_err = TRUE;
-    OBEX_SetUserCallBack(ctx->handle, idle_callback, NULL);
-    ctx->done = TRUE;
-    ctx->conid = CONID_INVALID;
-    if (ctx->conn_fd >= 0) {
-        OBEX_TransportDisconnect(ctx->handle);
-        close(ctx->conn_fd);
-        ctx->conn_fd = -1;
-    }
-    if (ctx->gio) {
-        g_io_channel_unref(ctx->gio);
-        ctx->gio = NULL;
-    }
-    if (ctx->gio_source) {
-        g_source_destroy(ctx->gio_source);
-        ctx->gio_source = NULL;
-    }
-    if (ctx->buf) {
-        /* Check that buffer is owned by us */
-        if (!(ctx->obex_op == OBEX_CMD_PUT && ctx->stream_fd < 0))
-            g_free(ctx->buf);
-        ctx->buf = NULL;
-        ctx->buf_size = 0;
-    }
-}
+static gboolean gw_obex_request_async(GwObex *ctx, obex_object_t *object) {
+    ctx->done = FALSE;
 
-static gboolean gw_obex_do_abort(GwObex *ctx) {
-    debug("gw_obex_do_abort()\n");
+    /* If this is a put, the first request can only be sent after we
+     * have some data in the outgoing buffer */
+    if (ctx->obex_op == OBEX_CMD_PUT) {
+        ctx->xfer->object = object;
+        return TRUE;
+    }
 
-    if (ctx->conn_fd < 0)
+    if (OBEX_Request(ctx->handle, object) < 0) {
+        debug("OBEX_Request() failed\n");
+        ctx->error = GW_OBEX_ERROR_INTERNAL;
         return FALSE;
+    }
 
-    ctx->abort = TRUE;
-
-#ifdef USE_NICE_ABORT
-    debug("Performing nice abort\n");
-    if (OBEX_CancelRequest(ctx->handle, TRUE) != 0)
-        return FALSE;
-    ctx->abort_sent = TRUE;
     return TRUE;
-#else
-    debug("Performing abort through disconnection (without ABORT command)\n");
-    ctx->done = TRUE;
-    OBEX_CancelRequest(ctx->handle, FALSE);
-    obex_link_error(ctx);
-    return FALSE;
-#endif
 }
 
 static gboolean gw_obex_request_sync(GwObex *ctx, obex_object_t *object) {
     /* Set sensible start values */
     ctx->done       = FALSE;
-    ctx->abort      = FALSE;
-    ctx->abort_sent = FALSE;
-    ctx->counter    = 0;
 
     if (OBEX_Request(ctx->handle, object) < 0) {
         debug("OBEX_Request() failed\n");
@@ -156,9 +108,9 @@ static gboolean gw_obex_request_sync(GwObex *ctx, obex_object_t *object) {
 
         if (ctx->cancel_cb && ctx->cancel_cb(ctx->cancel_data)) {
             debug("cancel_cb() returned TRUE, aborting.\n");
-            if (ctx->abort_sent)
+            if (ctx->xfer->abort)
                 continue;
-            if (!gw_obex_do_abort(ctx))
+            if (!gw_obex_xfer_do_abort(ctx->xfer))
                 break;
             /* Must continue to receive the abort reply */
             continue;
@@ -185,7 +137,7 @@ static gboolean gw_obex_request_sync(GwObex *ctx, obex_object_t *object) {
         /* It is possible that a EV_PROGRESS doesn't arrive after all data has
          * been transfered. Call pr_cb here to ensure app gets 100% progress */
         if (ctx->report_progress && ctx->pr_cb)
-            ctx->pr_cb(ctx, ctx->obex_op, ctx->counter, ctx->counter, ctx->pr_data);
+            ctx->pr_cb(ctx, ctx->obex_op, ctx->xfer->counter, ctx->xfer->counter, ctx->pr_data);
         return TRUE;
     }
     else
@@ -233,11 +185,11 @@ static void obex_connect_done(GwObex *ctx, obex_object_t *object, int obex_rsp)
         int new_size;
         debug("Version: 0x%02x. Flags: 0x%02x  OBEX packet length: %d\n",
                 nonhdrdata->version, nonhdrdata->flags, mtu);
+        /* Leave space for headers */
         new_size = mtu - 200;
-        /*OBEX_SetTransportMTU(ctx->handle, mtu, mtu);*/
-        if (new_size < ctx->chunk_size) {
+        if (new_size < ctx->tx_max) {
             debug("Resizing stream chunks to %d\n", new_size);
-            ctx->chunk_size = new_size;
+            ctx->tx_max = new_size;
         }
     }
 
@@ -311,12 +263,15 @@ static void show_headers(obex_t *handle, obex_object_t *object) {
 
     OBEX_ObjectReParseHeaders(handle, object);
 }
+#else
+static inline void show_headers(obex_t *handle, obex_object_t *object) {}
 #endif
 
 static void obex_abort_done(GwObex *ctx, obex_object_t *object,
                             int obex_cmd, int obex_rsp) {
     ctx->done = TRUE;
-    ctx->abort_sent = FALSE;
+    if (ctx->xfer)
+        ctx->xfer->do_cb = TRUE;
 
     if (obex_rsp != OBEX_RSP_SUCCESS)
         debug("ABORT of %s command (0x%02x) failed: %s (0x%02x)\n",
@@ -330,6 +285,9 @@ static void obex_abort_done(GwObex *ctx, obex_object_t *object,
 static void obex_request_done(GwObex *ctx, obex_object_t *object,
                               int obex_cmd, int obex_rsp) {
     ctx->done = TRUE;
+    if (ctx->xfer)
+        ctx->xfer->do_cb = TRUE;
+
     ctx->obex_rsp = obex_rsp;
 
     if (obex_rsp != OBEX_RSP_SUCCESS) {
@@ -383,27 +341,53 @@ static void get_target_size_and_time(obex_t *handle, obex_object_t *object,
 }
 
 static void obex_readstream(GwObex *ctx, obex_object_t *object) {
+    struct gw_obex_xfer *xfer = ctx->xfer;
     const uint8_t *buf;
     int actual;
 
-    if (ctx->counter == 0)
-        get_target_size_and_time(ctx->handle, object,
-                                 &ctx->target_size, &ctx->modtime);
+    if (!xfer) {
+        debug("Incomming data even though no xfer active!\n");
+        /* Flush incomming stream */
+        actual = OBEX_ObjectReadStream(ctx->handle, object, &buf);
+        if (actual > 0)
+            debug("Ignored %d bytes\n", actual);
+        return;
+    }
 
-#ifdef DEBUG
-    if (ctx->counter == 0)
+    if (ctx->xfer->counter == 0) {
+        get_target_size_and_time(ctx->handle, object,
+                                 &xfer->target_size, &xfer->modtime);
         show_headers(ctx->handle, object);
-#endif
+    }
     
     actual = OBEX_ObjectReadStream(ctx->handle, object, &buf);
     if (actual > 0) {
-        ctx->counter += actual;
-        if (ctx->stream_fd >= 0)
-            (void) write(ctx->stream_fd, buf, actual);
+        xfer->counter += actual;
+
+        debug("obex_readstream: got %d bytes (%d in total)\n", actual, xfer->counter);
+
+        if (xfer->async) {
+            gint free_space = xfer->buf_size - (xfer->data_start + xfer->data_length);
+            if (actual > free_space) {
+                /* This should never happen */
+                debug("Out of buffer space: actual=%d, free=%d\n", actual, free_space);
+                return;
+            }
+
+            memcpy(&xfer->buf[xfer->data_start], buf, actual);
+            xfer->data_length += actual;
+
+            debug("OBEX_SuspendRequest at %s:%d (%s)\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
+            OBEX_SuspendRequest(ctx->handle, object);
+
+            xfer->do_cb = TRUE;
+        }
+        else if (xfer->stream_fd >= 0)
+            (void) write(xfer->stream_fd, buf, actual);
         else {
-            ctx->buf = g_realloc(ctx->buf, ctx->counter);
-            memcpy(&ctx->buf[ctx->buf_size], buf, actual);
-            ctx->buf_size = ctx->counter;
+            xfer->buf = g_realloc(xfer->buf, xfer->counter);
+            memcpy(&xfer->buf[xfer->buf_size], buf, actual);
+            xfer->buf_size = xfer->counter;
         }
     }
     else
@@ -411,14 +395,49 @@ static void obex_readstream(GwObex *ctx, obex_object_t *object) {
 }
 
 static void obex_writestream(GwObex *ctx, obex_object_t *object) {
+    struct gw_obex_xfer *xfer = ctx->xfer;
     obex_headerdata_t hv;
     int actual = -1;
 
-    if (ctx->stream_fd >= 0) {
-        actual = read(ctx->stream_fd, ctx->buf, ctx->chunk_size);
-        hv.bs = ctx->buf;
+    if (!xfer) {
+        debug("Request to provide data even though no active xfer!");
+        hv.bs = NULL;
+        OBEX_ObjectAddHeader(ctx->handle, object, OBEX_HDR_BODY,
+                hv, 0, OBEX_FL_STREAM_DATAEND);
+        return;
+    }
+
+    if (xfer->async) {
+        if (xfer->data_length > 0) {
+            gint send_size = xfer->data_length > ctx->tx_max ? ctx->tx_max : xfer->data_length;
+
+            hv.bs = &xfer->buf[xfer->data_start];
+            OBEX_ObjectAddHeader(ctx->handle, object, OBEX_HDR_BODY,
+                    hv, send_size, OBEX_FL_STREAM_DATA);
+            actual = send_size;
+            xfer->data_length -= send_size;
+            if (xfer->data_length == 0)
+                xfer->data_start = 0;
+            else
+                xfer->data_start += send_size;
+
+            xfer->do_cb = TRUE;
+	    if (!xfer->close) {
+                debug("OBEX_SuspendRequest at %s:%d (%s)\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
+                OBEX_SuspendRequest(ctx->handle, object);
+            }
+        }
+        else {
+            hv.bs = NULL;
+            OBEX_ObjectAddHeader(ctx->handle, object, OBEX_HDR_BODY,
+                    hv, 0, OBEX_FL_STREAM_DATAEND);
+        }
+    }
+    else if (xfer->stream_fd >= 0) {
+        actual = read(xfer->stream_fd, xfer->buf, ctx->tx_max);
+        hv.bs = xfer->buf;
 #ifdef TEST_ABORT
-        if (ctx->counter > 4000)
+        if (xfer->counter > 4000)
             actual = -1;
 #endif
         if (actual > 0)
@@ -429,16 +448,16 @@ static void obex_writestream(GwObex *ctx, obex_object_t *object) {
                     hv, 0, OBEX_FL_STREAM_DATAEND);
         else { /* error reading file */
             debug("read(): %s\n", strerror(errno));
-            gw_obex_do_abort(ctx);
+            gw_obex_xfer_do_abort(xfer);
         }
     }
     else {
-        if (ctx->counter < ctx->buf_size) {
-            if (ctx->buf_size > ctx->counter + ctx->chunk_size)
-                actual = ctx->chunk_size;
+        if (xfer->counter < xfer->buf_size) {
+            if (xfer->buf_size > xfer->counter + ctx->tx_max)
+                actual = ctx->tx_max;
             else
-                actual = ctx->buf_size - ctx->counter;
-            hv.bs = &ctx->buf[ctx->counter]; 
+                actual = xfer->buf_size - xfer->counter;
+            hv.bs = &xfer->buf[xfer->counter]; 
             OBEX_ObjectAddHeader(ctx->handle, object, OBEX_HDR_BODY,
                     hv, actual, OBEX_FL_STREAM_DATA);
         }
@@ -450,7 +469,7 @@ static void obex_writestream(GwObex *ctx, obex_object_t *object) {
     }
 
     if (actual > 0)
-        ctx->counter += actual;
+        xfer->counter += actual;
 }
 
 static void obex_event_handler(obex_t *handle, obex_object_t *object, int mode,
@@ -464,7 +483,7 @@ static void obex_event_handler(obex_t *handle, obex_object_t *object, int mode,
         case OBEX_EV_PROGRESS:
             debug("OBEX_EV_PROGRESS\n");
             if (ctx->report_progress && ctx->pr_cb)
-                ctx->pr_cb(ctx, ctx->obex_op, ctx->counter, ctx->target_size, ctx->pr_data);
+                ctx->pr_cb(ctx, ctx->obex_op, ctx->xfer->counter, ctx->xfer->target_size, ctx->pr_data);
             break;
         case OBEX_EV_REQDONE:
             debug("OBEX_EV_REQDONE\n");
@@ -503,33 +522,55 @@ static void obex_event_handler(obex_t *handle, obex_object_t *object, int mode,
     }
 }
 
-gboolean fd_raw_mode(int fd) {
-    struct termios mode;
+gboolean gw_obex_set_error(GwObex *ctx) {
+    ctx->error = 0;
 
-    memset(&mode, 0, sizeof (mode));
-    if (tcgetattr(fd, &mode) < 0) {
-        debug("tcgetattr(%d, &mode): %s", fd, strerror(errno));
+    if (!ctx->done)
         return FALSE;
-    }
 
-    mode.c_iflag = 0;
-    mode.c_oflag &= ~OPOST;
-    mode.c_lflag &= ~(ISIG | ICANON | ECHO
-#ifdef XCASE
-            | XCASE
-#endif
-            );
-    mode.c_cc[VMIN] = 1;
-    mode.c_cc[VTIME] = 0;
+    if (ctx->xfer && ctx->xfer->abort)
+        ctx->error = GW_OBEX_ERROR_ABORT;
+    else if (ctx->conn_fd < 0 || ctx->link_err)
+        ctx->error = GW_OBEX_ERROR_DISCONNECT;
+    else
+        ctx->error = (gint)ctx->obex_rsp;
 
-    if (tcsetattr(fd, TCSADRAIN, &mode) < 0) {
-        debug("tcsetattr(%d, TCSADRAIN, &mode): %s", fd, strerror(errno));
+    if (ctx->error == OBEX_RSP_SUCCESS)
         return FALSE;
-    }
 
     return TRUE;
 }
 
+void obex_link_error(GwObex *ctx) {
+    if (ctx->link_err)
+        return;
+    ctx->link_err = TRUE;
+    OBEX_SetUserCallBack(ctx->handle, idle_callback, NULL);
+    ctx->done = TRUE;
+    ctx->conid = CONID_INVALID;
+    if (ctx->conn_fd >= 0) {
+        OBEX_TransportDisconnect(ctx->handle);
+        close(ctx->conn_fd);
+        ctx->conn_fd = -1;
+    }
+    if (ctx->gio) {
+        g_io_channel_unref(ctx->gio);
+        ctx->gio = NULL;
+    }
+    if (ctx->gio_source) {
+        g_source_destroy(ctx->gio_source);
+        ctx->gio_source = NULL;
+    }
+    if (ctx->xfer) {
+        /* Check that buffer is owned by us */
+        if (!(ctx->obex_op == OBEX_CMD_PUT && ctx->xfer->stream_fd < 0))
+            g_free(ctx->xfer->buf);
+        ctx->xfer->buf = NULL;
+        ctx->xfer->buf_size = 0;
+	ctx->xfer->do_cb = TRUE;
+    }
+}
+
 gboolean gw_obex_transport_setup(int fd, obex_t **handle) {
     *handle = OBEX_Init(OBEX_TRANS_FD, obex_event_handler, 0);
     if (*handle == NULL) {
@@ -549,7 +590,8 @@ gboolean gw_obex_transport_setup(int fd, obex_t **handle) {
 }
 
 void gw_obex_get_error(GwObex *ctx, gint *error) {
-    *error = ctx->error;
+    if (error)
+        *error = ctx->error;
     ctx->error = OBEX_RSP_SUCCESS;
 }
 
@@ -560,40 +602,27 @@ gboolean gw_obex_cb(GIOChannel *chan, GIOCondition cond, gpointer data) {
 
     GW_OBEX_LOCK(ctx);
 
-    if (ctx->conn_fd < 0) {
-        debug("obex_cb() called, but connection is closed\n");
+    if (ctx->conn_fd < 0 || (cond & (G_IO_ERR | G_IO_HUP | G_IO_NVAL))) {
+        debug("gw_obex_cb: error or connection closed\n");
         obex_link_error(ctx);
         GW_OBEX_UNLOCK(ctx);
+        if (ctx->xfer && ctx->xfer->cb)
+            ctx->xfer->cb(ctx->xfer, ctx->xfer->cb_data);
         if (ctx->dc_cb)
             ctx->dc_cb(ctx, ctx->dc_data);
         return FALSE;
     }
 
-    if (cond != G_IO_IN) {
-        switch (cond) {
-            case G_IO_ERR:
-                debug("Error on OBEX FD\n");
-                break;
-            case G_IO_HUP:
-                debug("HUP on OBEX FD\n");
-                break;
-            case G_IO_NVAL:
-                debug("OBEX FD was closed!\n");
-                break;
-            default:
-                debug("Unknown condition (%d) on conn_fd (%d)\n",
-                        cond, ctx->conn_fd);
-                //g_assert_not_reached();
-        }
-        obex_link_error(ctx);
-
-        GW_OBEX_UNLOCK(ctx);
-        if (ctx->dc_cb)
-            ctx->dc_cb(ctx, ctx->dc_data);
-        return FALSE;
-    }
-
+    debug("Calling OBEX_HandleInput\n");
     OBEX_HandleInput(ctx->handle, 0);
+    debug("Returned from OBEX_HandleInput\n");
+
+    if (ctx->xfer && ctx->xfer->cb && ctx->xfer->do_cb) {
+        ctx->xfer->do_cb = FALSE;
+        GW_OBEX_UNLOCK(ctx);
+        ctx->xfer->cb(ctx->xfer, ctx->xfer->cb_data);
+        GW_OBEX_LOCK(ctx);
+    }
 
     GW_OBEX_UNLOCK(ctx);
 
@@ -603,6 +632,11 @@ gboolean gw_obex_cb(GIOChannel *chan, GIOCondition cond, gpointer data) {
 gboolean gw_obex_disconnect(GwObex *ctx) {
     obex_object_t *object;
 
+    if (!ctx->done || ctx->xfer) {
+        ctx->error = GW_OBEX_ERROR_BUSY;
+        return FALSE;
+    }
+
     object = OBEX_ObjectNew(ctx->handle, OBEX_CMD_DISCONNECT);
 
     if (ctx->conid != CONID_INVALID) {
@@ -618,6 +652,11 @@ gboolean gw_obex_connect(GwObex *ctx, const char *target, size_t target_len) {
     gboolean ret;
     obex_object_t *object;
 
+    if (!ctx->done || ctx->xfer) {
+        ctx->error = GW_OBEX_ERROR_BUSY;
+        return FALSE;
+    }
+
     ctx->obex_op = OBEX_CMD_CONNECT;
 
     object = OBEX_ObjectNew(ctx->handle, OBEX_CMD_CONNECT);
@@ -640,18 +679,18 @@ GwObex *make_context(obex_t *handle) {
     context->handle      = handle;
     context->conn_fd     = OBEX_GetFD(handle);
     context->conid       = CONID_INVALID;
-    context->stream_fd   = -1;
-    context->chunk_size  = GW_OBEX_TX_MTU - 200;
+    context->tx_max      = GW_OBEX_TX_MTU - 200;
+    context->rx_max      = GW_OBEX_RX_MTU;
     context->obex_op     = OBEX_CMD_NONE;
     context->obex_rsp    = OBEX_RSP_SUCCESS;
-    context->target_size = GW_OBEX_UNKNOWN_LENGTH;
-    context->modtime     = -1;
+    context->done        = TRUE;
 
     return context;
 }
 
 gboolean gw_obex_action_op(GwObex *ctx, const gchar *src, const gchar *dst,
                            uint8_t action) {
+    gboolean ret = FALSE;
     obex_object_t *object;
     obex_headerdata_t hv;
     gunichar2 *uname;
@@ -659,6 +698,11 @@ gboolean gw_obex_action_op(GwObex *ctx, const gchar *src, const gchar *dst,
 
     g_assert(src && dst);
 
+    if (!ctx->done || ctx->xfer) {
+        ctx->error = GW_OBEX_ERROR_BUSY;
+        return FALSE;
+    }       
+
     ctx->obex_op = OBEX_CMD_ACTION;
 
     object = OBEX_ObjectNew(ctx->handle, OBEX_CMD_ACTION);
@@ -674,7 +718,7 @@ gboolean gw_obex_action_op(GwObex *ctx, const gchar *src, const gchar *dst,
     uname_len = get_uname(&uname, src);
     if (uname_len < 0) {
         OBEX_ObjectDelete(ctx->handle, object);
-        return FALSE;
+        goto out;
     }
     hv.bs = (unsigned char *)uname;
     OBEX_ObjectAddHeader(ctx->handle, object, OBEX_HDR_NAME, hv, uname_len, 0);
@@ -683,45 +727,49 @@ gboolean gw_obex_action_op(GwObex *ctx, const gchar *src, const gchar *dst,
     uname_len = get_uname(&uname, dst);
     if (uname_len < 0) {
         OBEX_ObjectDelete(ctx->handle, object);
-        return FALSE;
+        goto out;
     }
     hv.bs = (unsigned char *)uname;
     OBEX_ObjectAddHeader(ctx->handle, object, OBEX_HDR_DESTNAME, hv, uname_len, 0);
     g_free(uname);
 
-    ctx->obex_op = OBEX_CMD_NONE;
+    ret = gw_obex_request_sync(ctx, object);
 
-    return gw_obex_request_sync(ctx, object);
+out:
+    ctx->obex_op = OBEX_CMD_NONE;
+    return ret;
 }
 
 gboolean gw_obex_setpath(GwObex *ctx, const gchar *path, int flags) {
-    gboolean ret, is_name;
+    gboolean ret = FALSE;
     obex_headerdata_t hv;
     obex_object_t *object;
     obex_setpath_hdr_t nonhdrdata;
     gunichar2 *uname; 
     glong uname_len;
 
+    if (!ctx->done || ctx->xfer) {
+        ctx->error = GW_OBEX_ERROR_BUSY;
+        return FALSE;
+    }       
+
     ctx->obex_op = OBEX_CMD_SETPATH;
 
     nonhdrdata.flags = 0x02;
     nonhdrdata.constants = 0;
 
-    if (path == NULL || *path == '\0') {
-        /* empty name. change to root directory */
-        uname = (gunichar2 *)g_strdup("");
-        uname_len = 0;
-        is_name = TRUE;
-    } else if (strcmp(path, "..") == 0) {
+    if (strcmp(path, "..") == 0) {
         /* move up one directory */
         nonhdrdata.flags = 0x03;
-        is_name = FALSE;
-        uname = NULL;
-        uname_len = 0;
-    } else {
+        uname_len = -1;
+    }
+    else {
         /* normal directory change */
         uname_len = get_uname(&uname, path);
-        is_name = TRUE;
+        if (uname_len < 0) {
+            ctx->error = GW_OBEX_ERROR_INVALID_PARAMS;
+            goto out;
+        }
     }
 
     if (flags & SETPATH_CREATE)
@@ -735,43 +783,40 @@ gboolean gw_obex_setpath(GwObex *ctx, const gchar *path, int flags) {
         OBEX_ObjectAddHeader(ctx->handle, object, OBEX_HDR_CONNECTION, hv, 4, 0);
     }
 
-    if (is_name) {
-        if (uname_len < 0) {
-            OBEX_ObjectDelete(ctx->handle, object);
-            return FALSE;
-        }
-        hv.bs = (unsigned char *)uname;
+    if (uname_len >= 0) {
+        hv.bs = (unsigned char *) (uname ? (char *)uname : "");
         OBEX_ObjectAddHeader(ctx->handle, object, OBEX_HDR_NAME, hv, uname_len, 0);
+        g_free(uname);
     }
 
     ret = gw_obex_request_sync(ctx, object);
-    if (uname)
-        g_free(uname);
     
+out:
     ctx->obex_op = OBEX_CMD_NONE;
-    
     return ret;
 }
 
 gboolean gw_obex_get(GwObex *ctx,
                      const gchar *local, const gchar *remote, const gchar *type,
-                     gchar **buf, gint *buf_size) {
-    gboolean ret;
-    gboolean user_fd = FALSE;
+                     gchar **buf, gint *buf_size, int stream_fd,
+                     gboolean async) {
+    gboolean ret = FALSE;
     obex_headerdata_t hv;
     obex_object_t *object;
-#ifdef DEBUG
-    time_t tm;
-#endif
 
-    g_assert(local || buf || ctx->stream_fd > 0);
+    g_assert(local || buf || stream_fd > 0 || async);
     g_assert(remote || type);
 
+    if (!ctx->done || ctx->xfer) {
+        ctx->error = GW_OBEX_ERROR_BUSY;
+        return ret;
+    }       
+
     ctx->obex_op = OBEX_CMD_GET;
 
-    object = OBEX_ObjectNew(ctx->handle, OBEX_CMD_GET);
+    ctx->xfer = gw_obex_xfer_new(ctx, async);
 
-    ctx->target_size = GW_OBEX_UNKNOWN_LENGTH;
+    object = OBEX_ObjectNew(ctx->handle, OBEX_CMD_GET);
 
     if (ctx->conid != CONID_INVALID) {
         hv.bq4 = ctx->conid;
@@ -783,102 +828,80 @@ gboolean gw_obex_get(GwObex *ctx,
         OBEX_ObjectAddHeader(ctx->handle, object, OBEX_HDR_TYPE, hv, strlen(type) + 1, 0);
     }
 
-    if (ctx->stream_fd >= 0)
-        user_fd = TRUE;
-
     if (remote) {
         gunichar2 *uname;
         glong uname_len;
+
         uname_len = get_uname(&uname, remote);
         if (uname_len < 0) {
             OBEX_ObjectDelete(ctx->handle, object);
             ctx->error = GW_OBEX_ERROR_INVALID_PARAMS;
-            return FALSE;
+            goto out;
         }
-        hv.bs = (unsigned char *)uname;
+
+        /* OpenOBEX is buggy and won't append the header unless hv.bs != NULL */
+        hv.bs = (unsigned char *) (uname ? (char *)uname : "");
+
         OBEX_ObjectAddHeader(ctx->handle, object, OBEX_HDR_NAME, hv, uname_len, 0);
         g_free(uname);
     }
 
     if (local) {
-        ctx->stream_fd = open(local, O_WRONLY | O_CREAT, 0600);
-        if (ctx->stream_fd < 0) {
+        ctx->xfer->stream_fd = open(local, O_WRONLY | O_CREAT, 0600);
+        if (ctx->xfer->stream_fd < 0) {
             if (errno == ENOENT || errno == ENODEV)
                 ctx->error = GW_OBEX_ERROR_INVALID_PARAMS;
             else
                 ctx->error = GW_OBEX_ERROR_LOCAL_ACCESS;
             debug("open(%s): %s", local, strerror(errno));
             OBEX_ObjectDelete(ctx->handle, object);
-            return FALSE;
+            goto out;
         }
     }
-    else if (!user_fd) {
-        g_assert(buf && buf_size);
-        ctx->buf = NULL;
-        ctx->buf_size = 0;
-        ctx->stream_fd = -1;
-    }
-
-    ctx->report_progress = TRUE;
 
     OBEX_ObjectReadStream(ctx->handle, object, NULL);
 
-#ifdef DEBUG
-    tm = time(NULL);
-#endif
-    ret = gw_obex_request_sync(ctx, object);
-    if (ctx->stream_fd >= 0) {
-        if (!user_fd)
-            close(ctx->stream_fd);
-        ctx->stream_fd = -1;
+    if (async) {
+        ret = gw_obex_request_async(ctx, object);
+        if (ret)
+            return ret;
     }
+    else {
+        ctx->report_progress = TRUE;
+        ret = gw_obex_request_sync(ctx, object);
+    }
+
+    if (ctx->xfer->stream_fd >= 0 && stream_fd < 0)
+        close(ctx->xfer->stream_fd);
+
     if (ret == FALSE) {
         if (local)
             unlink(local);
-        if (ctx->buf) {
-            g_free(ctx->buf);
-            ctx->buf = NULL;
-            ctx->buf_size = 0;
-        }
     }
     else {
         if (local) {
             debug("%s stored in %s\n", remote ? remote : type, local);
-            if (ctx->modtime != -1) {
+            if (ctx->xfer->modtime != -1) {
                 struct utimbuf ubuf;
                 ubuf.actime = time(NULL);
-                ubuf.modtime = ctx->modtime;
+                ubuf.modtime = ctx->xfer->modtime;
                 if (utime(local, &ubuf) < 0)
                     debug("utime(%s): %s\n", local, g_strerror(errno));
             }
         }
         if (buf) {
-            *buf = (gchar *)ctx->buf;
-            *buf_size = ctx->buf_size;
-            ctx->target_size = ctx->buf_size;
-            ctx->buf = NULL;
-            ctx->buf_size = 0;
+            *buf = (gchar *)ctx->xfer->buf;
+            *buf_size = ctx->xfer->buf_size;
+            /* Make sure gw_obex_xfer_free doesn't free the buffer */
+            ctx->xfer->buf = NULL;
         }
     }
 
-#ifdef DEBUG
-    tm = time(NULL) - tm;
-    if (tm > 0 && ctx->target_size > 0) {
-        double speed = (double)ctx->target_size / tm;
-        if (speed < 1000)
-            debug("transfer speed: %.2f byte/s\n", speed);
-        else
-            debug("transfer speed: %.2f kB/s\n", speed / 1024);
-    }
-    else if (ctx->target_size > 0) {
-        debug("%d bytes transfered in less than one second\n", ctx->target_size);
-    }
-#endif
+out:
+    _gw_obex_xfer_free(ctx->xfer);
+    ctx->xfer = NULL;
 
     ctx->report_progress = FALSE;
-    ctx->target_size = GW_OBEX_UNKNOWN_LENGTH;
-    ctx->modtime     = -1;
-
     ctx->obex_op = OBEX_CMD_NONE;
 
     return ret;
@@ -886,74 +909,62 @@ gboolean gw_obex_get(GwObex *ctx,
 
 gboolean gw_obex_put(GwObex *ctx,
                      const gchar *local, const gchar *remote, const gchar *type,
-                     const gchar *buf, gint buf_size) {
-    gboolean ret;
-    gboolean user_fd = FALSE;
+                     const gchar *buf, gint object_size, time_t object_time,
+                     int stream_fd, gboolean async) {
+    gboolean ret = FALSE;
     obex_headerdata_t hv;
     obex_object_t *object;
     gunichar2 *uname = NULL;
     glong uname_len = 0;
-#ifdef DEBUG
-    time_t tm;
-#endif
 
     g_assert(remote || type);
 
+    if (!ctx->done || ctx->xfer) {
+        ctx->error = GW_OBEX_ERROR_BUSY;
+        return FALSE;
+    }
+
     ctx->obex_op = OBEX_CMD_PUT;
 
     if (remote) {
         uname_len = get_uname(&uname, remote);
         if (uname_len < 0) {
             ctx->error = GW_OBEX_ERROR_INVALID_PARAMS;
-            return FALSE;
+            goto out;
         }
     }
 
-    if (ctx->stream_fd >= 0)
-        user_fd = TRUE;
-
-    ctx->report_progress = TRUE;
+    ctx->xfer = gw_obex_xfer_new(ctx, async);
 
     if (local) {
         if (file_is_dir(local)) {
             debug("Trying to PUT a directory\n");
-            if (uname)
-                g_free(uname);
             ctx->error = GW_OBEX_ERROR_INVALID_PARAMS;
-            return FALSE;
+            goto out;
         }
 
-        ctx->stream_fd = open(local, O_RDONLY);
-        if (ctx->stream_fd < 0) {
+        ctx->xfer->stream_fd = open(local, O_RDONLY);
+        if (ctx->xfer->stream_fd < 0) {
             if (errno == ENOENT || errno == ENODEV)
                 ctx->error = GW_OBEX_ERROR_INVALID_PARAMS;
             else
                 ctx->error = GW_OBEX_ERROR_LOCAL_ACCESS;
             debug("open(%s): %s", local, strerror(errno));
-            if (uname)
-                g_free(uname);
-            return FALSE;
+            goto out;
         }
-        g_assert(ctx->buf == NULL);
-        ctx->buf = g_malloc(ctx->chunk_size);
-        ctx->buf_size = ctx->chunk_size;
+        ctx->xfer->buf = g_malloc(ctx->tx_max);
+        ctx->xfer->buf_size = ctx->tx_max;
         debug("Sending %s to %s\n", local, remote ? remote : type);
     }
     else if (buf) {
-        ctx->buf = (unsigned char *)buf;
-        ctx->buf_size = buf_size;
-        ctx->stream_fd = -1;
+        ctx->xfer->buf = (unsigned char *)buf;
+        ctx->xfer->buf_size = object_size;
         debug("Sending to %s\n", remote ? remote : type);
     }
-    else if (user_fd) {
-        ctx->buf = NULL;
-        ctx->buf_size = 0;
-    }
-    else { /* Delete */
+    else if (stream_fd >= 0)
+        ctx->xfer->stream_fd = stream_fd;
+    else if (!async) { /* Delete */
         ctx->report_progress = FALSE;
-        ctx->stream_fd = -1;
-        ctx->buf = NULL;
-        ctx->buf_size = 0;
         debug("Deleting %s\n", remote ? remote : type);
     }
 
@@ -968,6 +979,7 @@ gboolean gw_obex_put(GwObex *ctx,
         hv.bs = (unsigned char *)uname;
         OBEX_ObjectAddHeader(ctx->handle, object, OBEX_HDR_NAME, hv, uname_len, 0);
         g_free(uname);
+        uname = NULL;
     }
 
     if (type) {
@@ -975,80 +987,65 @@ gboolean gw_obex_put(GwObex *ctx,
         OBEX_ObjectAddHeader(ctx->handle, object, OBEX_HDR_TYPE, hv, strlen(type) + 1, 0);
     }
 
-    if (ctx->stream_fd >= 0 || buf) {
-        gint size = -1;
-        time_t time = -1;
-
-        if (buf)
-            size = buf_size;
-        else {
-            struct stat stats;
-            if (fstat(ctx->stream_fd, &stats) == 0) {
-                size = stats.st_size;
-                time = stats.st_mtime;
-            }
+    /* Try to figure out modification time if none was given */
+    if (ctx->xfer->stream_fd >= 0) {
+        struct stat stats;
+        if (fstat(ctx->xfer->stream_fd, &stats) == 0) {
+            object_size = stats.st_size;
+            if (object_time < 0)
+                object_time = stats.st_mtime;
         }
+    }
 
-        if (time >= 0) {
-            char tstr[17];
-            int len;
+    /* Add a time header if possible */
+    if (object_time >= 0) {
+        char tstr[17];
+        int len;
 
-            len = make_iso8601(time, tstr, sizeof(tstr));
+        len = make_iso8601(object_time, tstr, sizeof(tstr));
 
-            if (len >= 0) {
-                debug("Adding time header: %s\n", tstr);
-                hv.bs = (unsigned char *)tstr;
-                OBEX_ObjectAddHeader(ctx->handle, object, OBEX_HDR_TIME, hv, len, 0);
-            }
+        if (len >= 0) {
+            debug("Adding time header: %s\n", tstr);
+            hv.bs = (unsigned char *)tstr;
+            OBEX_ObjectAddHeader(ctx->handle, object, OBEX_HDR_TIME, hv, len, 0);
         }
+    }
 
-        if (size >= 0) {
-            ctx->target_size = size;
-            debug("Adding size header: %d\n", size);
-            hv.bq4 = (uint32_t)size;
-            OBEX_ObjectAddHeader(ctx->handle, object, OBEX_HDR_LENGTH, hv, 4, 0);
-        }
-        else
-           ctx->target_size = GW_OBEX_UNKNOWN_LENGTH; 
+    /* Add a length header if possible */
+    if (object_size > 0) {
+        ctx->xfer->target_size = object_size;
+        debug("Adding size header: %d\n", object_size);
+        hv.bq4 = (uint32_t)object_size;
+        OBEX_ObjectAddHeader(ctx->handle, object, OBEX_HDR_LENGTH, hv, 4, 0);
+    }
+    else
+        ctx->xfer->target_size = GW_OBEX_UNKNOWN_LENGTH; 
 
+    if (ctx->xfer->stream_fd >= 0 || buf || async) {
         hv.bs = NULL;
         OBEX_ObjectAddHeader(ctx->handle, object, OBEX_HDR_BODY, hv, 0, OBEX_FL_STREAM_START);
     }
 
-#ifdef DEBUG
-    tm = time(NULL);
-#endif
-    ret = gw_obex_request_sync(ctx, object);
+    if (async) {
+        ret = gw_obex_request_async(ctx, object);
+        if (ret)
+            return ret;
+    }
+    else {
+        ctx->report_progress = TRUE;
+        ret = gw_obex_request_sync(ctx, object);
+    }
 
-    if (ctx->stream_fd >= 0) {
-        if (!user_fd)
-            close(ctx->stream_fd);
-        ctx->stream_fd = -1;
-        if (ctx->buf) {
-            g_free(ctx->buf);
-            ctx->buf = NULL;
-        }
-    }
-#ifdef DEBUG
-    if (ret == TRUE) {
-        tm = time(NULL) - tm;
-        if (tm > 0 && ctx->target_size > 0) {
-            double speed = (double)ctx->target_size / tm;
-            if (speed < 1024)
-                debug("transfer speed: %.2f byte/s\n", speed);
-            else
-                debug("transfer speed: %.2f kB/s\n", speed / 1024);
-        }
-        else if (ctx->target_size > 0)
-            debug("%d bytes transfered in less than one second\n", ctx->target_size);
-    }
-#endif
+out:
+    g_free(uname);
+
+    if (ctx->xfer->stream_fd >= 0 && stream_fd < 0)
+        close(ctx->xfer->stream_fd);
+
+    _gw_obex_xfer_free(ctx->xfer);
+    ctx->xfer = NULL;
 
     ctx->report_progress = FALSE;
-    ctx->buf = NULL;
-    ctx->buf_size = 0;
-    ctx->target_size = GW_OBEX_UNKNOWN_LENGTH;
-
     ctx->obex_op = OBEX_CMD_NONE;
 
     return ret;
