@@ -24,12 +24,29 @@
 #ifndef _OBEX_PRIV_H_
 #define _OBEX_PRIV_H_
 
+#include <stdio.h>
 #include <stdint.h>
 #include <time.h>
 #include <glib.h>
 
 #include <openobex/obex.h>
 
+#ifdef HAVE_CONFIG_H
+# include "config.h"
+#endif
+
+#include "gw-obex.h"
+#include "obex-xfer.h"
+
+#define CHECK_DISCONNECT(ret,err,ctx) do { \
+                                          if ((ctx)->conn_fd < 0) { \
+                                              if (err) \
+                                                  *(err) = GW_OBEX_ERROR_DISCONNECT; \
+                                              GW_OBEX_UNLOCK(ctx); \
+                                              return (ret); \
+                                          } \
+                                      } while (0)
+
 #ifndef OBEX_CMD_ACTION
 # define OBEX_CMD_ACTION      0x06
 
@@ -57,8 +74,21 @@
 #define LST_TYPE "x-obex/folder-listing"
 
 #ifdef G_THREADS_ENABLED
-# define GW_OBEX_LOCK(ctx)   g_mutex_lock((ctx)->mutex)
-# define GW_OBEX_UNLOCK(ctx) g_mutex_unlock((ctx)->mutex)
+# ifdef DEBUG
+#  define GW_OBEX_LOCK(ctx) do { \
+      debug("Attempting GW_OBEX_LOCK at %s:%d (%s)...", __FILE__, __LINE__, __PRETTY_FUNCTION__); \
+      fflush(stdout); \
+      g_mutex_lock((ctx)->mutex); \
+      debug("got it!\n"); \
+   } while (0)
+#  define GW_OBEX_UNLOCK(ctx) do { \
+      debug("Unlocking GW_OBEX_LOCK at %s:%d (%s)\n", __FILE__, __LINE__, __PRETTY_FUNCTION__); \
+      g_mutex_unlock((ctx)->mutex); \
+   } while (0)
+# else
+#  define GW_OBEX_LOCK(ctx)   g_mutex_lock((ctx)->mutex)
+#  define GW_OBEX_UNLOCK(ctx) g_mutex_unlock((ctx)->mutex)
+# endif
 #else
 # define GW_OBEX_LOCK(ctx)   ((void)(0))
 # define GW_OBEX_UNLOCK(ctx) ((void)(0))
@@ -92,6 +122,9 @@ struct gw_obex {
     gw_obex_progress_cb_t    pr_cb;
     gpointer                 pr_data;
 
+    /* Whether calling pr_cb is necessary or not */
+    gboolean                 report_progress;
+
     /* Cancel callback and associated data */
     gw_obex_cancel_cb_t      cancel_cb;
     gpointer                 cancel_data;
@@ -99,21 +132,14 @@ struct gw_obex {
     /* For checking if the current operation is finished */
     gboolean                 done;
     
-    /* TRUE if the current operation was aborted */
-    gboolean                 abort;
-
-    /* TRUE if the ABORT command has been sent by us */
-    gboolean                 abort_sent;
-
     /* TRUE if a link error has hapened */
     gboolean                 link_err;
 
-    /* When doing a get or put for a local file */
-    int                      stream_fd;
-
     /* FD for the transport connection */
     int                      conn_fd;
 
+    GMainContext            *main_ctx;
+
     /* The transport connection's GIOChannel */
     GIOChannel               *gio;
 
@@ -129,33 +155,27 @@ struct gw_obex {
     /* The current OBEX operation */
     uint8_t                  obex_op;
 
-    /* Temporary buffer when doing a put or get */
-    unsigned char           *buf;
-    size_t                   buf_size;
-
-    /* Bytes read or written for the current get/put operation */
-    size_t                   counter;
-
-    /* Bytes to read at a time when doing a put */
-    uint16_t                 chunk_size;
-
-    /* Target length of the current get/put operation */
-    gint                     target_size;
-
     /* This is set if some operation fails */
     gint                     error;
 
-    /* Modification time of last file transfered */
-    time_t                   modtime;
+    /* Bytes to read at a time when doing a put */
+    uint16_t                 tx_max;
 
-    /* Whether calling pr_cb is necessary or not */
-    gboolean                 report_progress;
+    /* How many bytes to allocate for incomming object data */
+    uint16_t                 rx_max;
+
+    /* Current object transfer handle */
+    struct gw_obex_xfer     *xfer;
 };
 
 GwObex *make_context(obex_t *handle);
 
+gboolean gw_obex_set_error(GwObex *ctx);
+
 void gw_obex_get_error(GwObex *ctx, gint *error);
 
+void obex_link_error(GwObex *ctx);
+
 gboolean gw_obex_cb(GIOChannel *chan, GIOCondition cond, gpointer data);
 
 gboolean gw_obex_connect(GwObex *ctx, const char *target, size_t target_len);
@@ -169,7 +189,7 @@ gboolean gw_obex_action_op(GwObex *ctx, const gchar *src, const gchar *dst,
 
 gboolean gw_obex_setpath(GwObex *ctx, const gchar *path, int flags);
 
-/** Get an object from the GW
+/** Get an object from the server
  * @param ctx Pointer returned by gw_obex_setup()
  * @param local Local filename which contains the object
  * @param remote Remote filename to store the object in
@@ -178,9 +198,10 @@ gboolean gw_obex_setpath(GwObex *ctx, const gchar *path, int flags);
  */
 gboolean gw_obex_get(GwObex *ctx,
                      const gchar *local, const gchar *remote, const gchar *type,
-                     gchar **buf, gint *buf_size);
+                     gchar **buf, gint *buf_size, int stream_fd,
+                     gboolean async);
 
-/** Send an object to the GW
+/** Send an object to the server
  * @param ctx Pointer returned by gw_obex_setup()
  * @param local Local filename to store the objec in
  * @param remote Remote filename which contains the object
@@ -189,8 +210,7 @@ gboolean gw_obex_get(GwObex *ctx,
  */
 gboolean gw_obex_put(GwObex *ctx,
                      const gchar *local, const gchar *remote, const gchar *type,
-                     const gchar *buf, gint buf_size);
-
-gboolean fd_raw_mode(int fd);
+                     const gchar *buf, gint buf_size, time_t object_time,
+                     int stream_fd, gboolean async);
 
 #endif /* _OBEX_PRIV_H_ */
