@@ -23,9 +23,11 @@
 */
 #include <stdlib.h>
 #include <stdio.h>
+#include <errno.h>
 #include <string.h>
 #include <time.h>
 #include <glib.h>
+#include <termios.h>
 
 #ifdef HAVE_CONFIG_H
 # include "config.h"
@@ -46,9 +48,41 @@ char *bytestr(const uint8_t *uuid, int len) {
 }
 #endif
 
+gboolean fd_raw_mode(int fd) {
+    struct termios mode;
+
+    memset(&mode, 0, sizeof (mode));
+    if (tcgetattr(fd, &mode) < 0) {
+        debug("tcgetattr(%d, &mode): %s", fd, strerror(errno));
+        return FALSE;
+    }
+
+    mode.c_iflag = 0;
+    mode.c_oflag &= ~OPOST;
+    mode.c_lflag &= ~(ISIG | ICANON | ECHO
+#ifdef XCASE
+            | XCASE
+#endif
+            );
+    mode.c_cc[VMIN] = 1;
+    mode.c_cc[VTIME] = 0;
+
+    if (tcsetattr(fd, TCSADRAIN, &mode) < 0) {
+        debug("tcsetattr(%d, TCSADRAIN, &mode): %s", fd, strerror(errno));
+        return FALSE;
+    }
+
+    return TRUE;
+}
+
 glong get_uname(gunichar2 **uname, const gchar *name) {
     glong uname_len;
 
+    if (*name == '\0') {
+        *uname = NULL;
+        return 0;
+    }
+
     *uname = g_utf8_to_utf16(name, -1, NULL, &uname_len, NULL);
 
     if (*uname == NULL)
